/* punch (c) 2004 Jim Ursetto */

/* Polygon testing harness.  Benchmarks rendering time of large opaque,
 * translucent, and punch-through polys.
 * Press A to advance to the next test phase.
 * Press B to toggle textures on and off.
 * Press START to exit. */

#include <kos.h>
#include <math.h>
#include <plx/texture.h>
#include <plx/context.h>
#include <plx/prim.h>

extern uint8 romdisk[];
KOS_INIT_FLAGS(INIT_DEFAULT | INIT_MALLOCSTATS);
KOS_INIT_ROMDISK(romdisk);


static inline void draw_square(uint32 color) {
	plx_dr_state_t dr;
	float x=20,y=40,dx=600,dy=400;
	static float z=1.0f;

	plx_dr_init(&dr);

	plx_vert_ifd(&dr, PLX_VERT, x,    y+dy, z, color, 0.0f, 1.0f);
	plx_vert_ifd(&dr, PLX_VERT, x,    y   , z, color, 0.0f, 0.0f);
	plx_vert_ifd(&dr, PLX_VERT, x+dx, y+dy, z, color, 1.0f, 1.0f);
	plx_vert_ifd(&dr, PLX_VERT_EOS, x+dx, y, z, color, 1.0f, 0.0f);
	z+=0.1f;
}

/* store current and previous controller state */
typedef struct { 
	cont_state_t *st; 
	cont_state_t st_old;
} cont_persist_t;
#define BUTTON_PRESSED(persist, btn) (((persist.st->buttons & btn) && !(persist.st_old.buttons & btn)))

/* The various test phases, and their data: number of polys to draw, and
 * description.  By default, the polys are set to the maximum drawable in one
 * frame with textures on, except for PHASE_PT_OP, which is calibrated for
 * textures off.  */
enum { PHASE_OP, PHASE_TR_OP, PHASE_TR_TR, PHASE_TR_TR_OBSCURED, PHASE_PT_OP,
	PHASE_PT_OP_OVER, PHASE_PT_TR, PHASE_END };

struct {
	int polys;  /* polys to draw */
	char *desc;
} phase_data[] = { 
	{ 102, "Opaque list" },
	{   5, "Translucent list, opaque polygons" },
	{   5, "Translucent list, translucent polygons" },
	{  97, "Translucent list, translucent polygons, obscured by opaque poly" },
	{  17, "Punch-through list, opaque polygons" }, /* max 11 if textured */
	{  10, "Punch-through list, opaque polygons layered over translucent" },
	{   4, "Punch-through list, translucent (invisible) polygons" }, /* 4 if textured */
	{   0, "End" },
};

/* Run through all the phases.  Change this to alter the order or number of tests. */
int phases[] = { PHASE_OP, PHASE_TR_OP, PHASE_TR_TR, PHASE_TR_TR_OBSCURED, PHASE_PT_OP, 
                 PHASE_PT_OP_OVER, PHASE_PT_TR, PHASE_END };
/* int phases[] = { PHASE_OP, PHASE_TR_TR, PHASE_PT_OP_OVER, PHASE_END }; */

pvr_init_params_t pvr_params = {
	    /* Enable opaque and translucent polygons with size 16 */
		/* Also enable punch-through polygons -- binsize 8 is arbitrary */
	    { PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_8 },
	    /* Vertex buffer size */
	    512*1024
};

int main(int argc, char **argv) {
	int done, i;
	int phase = 0;
	int polys = -1;
	pvr_stats_t stats;
	cont_persist_t cstate;
	maple_device_t *dev;
	plx_texture_t *punch_txr;
	int txr_enabled = 1;

	cont_btn_callback(0, CONT_START | CONT_A, (cont_btn_callback_t)arch_exit);

	// Init PVR
	pvr_init(&pvr_params);
	pvr_set_bg_color(0.1f, 0.4f, 0.2f);

	// Load our texture
	punch_txr = plx_txr_load("/rd/punch.png", 1, 0);
	if (!punch_txr) return -1;
	/* It's not currently possible to get an ARGB1555 image from libpng.
	 * However, for our white-and-transparent image, 1555 and 4444 are 
	 * effectively the same.  There's no performance difference between
	 * 4444 and 1555 for punch-throughs; but note, I didn't check if the 
	 * PVR treats every mode as 1555. */
	// punch_txr->cxt_opaque.txr.format = PVR_TXRFMT_ARGB1555;
	plx_txr_setfilter(punch_txr, PVR_FILTER_NONE);

	// Setup the context
	plx_cxt_init();
	plx_cxt_texture(punch_txr);
	plx_cxt_culling(PLX_CULL_NONE);

	polys = phase_data[phases[phase]].polys;
	printf("phase: %s (%d polys)\n", phase_data[phases[phase]].desc, polys);

	for (done = 0; !done; ) {
		/* Read first available controller */
		dev = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
		if (!dev) done = 1;
		cstate.st = (cont_state_t *)maple_dev_status(dev);
		if (BUTTON_PRESSED(cstate, CONT_START))
			done = 1;
		if (BUTTON_PRESSED(cstate, CONT_B)) {
			txr_enabled ^= 1;
			printf("\ntextures %s\n", txr_enabled ? "enabled" : "disabled");
		}
		if (BUTTON_PRESSED(cstate, CONT_A)) {
			/* A: dump statistics, go to next phase */
			pvr_get_stats(&stats);
			dbglog(DBG_DEBUG, "\n3D Stats: %ld vblanks, frame rate ~%f fps, max vertex used %d bytes\n",
					stats.vbl_count, (double)stats.frame_rate, stats.vtx_buffer_used_max);
			pvr_reset_stats();
			phase++;
			polys = phase_data[phases[phase]].polys;
			printf("\nphase: %s (%d polys)\n", phase_data[phases[phase]].desc, polys);
		}
		memcpy(&cstate.st_old, cstate.st, sizeof(cstate.st_old));

		// Setup the frame
		vid_border_color(255,0,0);
		pvr_wait_ready();

		/* Stats will be wrong once after being reset, because render completion
		 * sets the last_time field to (current - start), which will be current - -1. */
		pvr_get_stats(&stats);
		printf("%d ", stats.rnd_last_time);

		pvr_scene_begin();
		vid_border_color(0,255,0);

		plx_cxt_texture(NULL);

		/* The horrible meat of the program */
		switch (phases[phase]) {
			case PHASE_OP:
				pvr_list_begin(PVR_LIST_OP_POLY);
				if (txr_enabled) plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_OP_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0xff000080);
				break;
			case PHASE_TR_OP:
				pvr_list_begin(PVR_LIST_TR_POLY);
				if (txr_enabled) plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_TR_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0xff000080);
				break;
			case PHASE_TR_TR:
				pvr_list_begin(PVR_LIST_TR_POLY);
				if (txr_enabled) plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_TR_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0x50000080);
				break;
			case PHASE_TR_TR_OBSCURED:
				pvr_list_begin(PVR_LIST_TR_POLY);
				plx_cxt_send(PVR_LIST_TR_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0x50000080);
				/* put an opaque (textured) poly in front */
				pvr_list_begin(PVR_LIST_OP_POLY);
				if (txr_enabled) plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_OP_POLY);
				draw_square(0xff000080);
				break;
			case PHASE_PT_OP:
				pvr_list_begin(PVR_LIST_PT_POLY);
				if (txr_enabled) plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_PT_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0xff000080);
				break;
			case PHASE_PT_OP_OVER:
				/* Draw the punch out polys over translucent polys */
				pvr_list_begin(PVR_LIST_TR_POLY);
				plx_cxt_texture(NULL);
				plx_cxt_send(PVR_LIST_TR_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0x50777777);
				pvr_list_begin(PVR_LIST_PT_POLY);
				if (txr_enabled)
					plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_PT_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0xffffffff);
				break;
			case PHASE_PT_TR:
				pvr_list_begin(PVR_LIST_PT_POLY);
				plx_cxt_texture(punch_txr);
				plx_cxt_send(PVR_LIST_PT_POLY);
				for (i = 0; i < polys; i++)
					draw_square(0xe0000080); /* slightly translucent (invisible in this case) */
				break;
			case PHASE_END:
				done = 1;
				break;

		}

		vid_border_color(255,255,0);
		pvr_scene_finish();

	}

	return 0;
}

